<?php
	/*
	* 2018 - VEGA Informatique
	*
	* NOTICE OF LICENSE
	*
	* This source file is subject to the Open Software License (OSL 3.0)
	* that is bundled with this package in the file LICENSE.txt.
	* It is also available through the world-wide-web at this URL:
	* http://opensource.org/licenses/osl-3.0.php
	*
	* DISCLAIMER
	*
	* Do not edit or add to this file if you wish to upgrade PolarisLink to newer
	* versions in the future. 
	*
	*  @author VEGA Informatique <contact@vega-info.fr>
	*  @copyright  2018 VEGA Informatique
	*  @license    http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
	*/

	require_once(_PS_ROOT_DIR_ . "/modules/polarislink/classes/PolarisLinkImportBase.php");
	
	/*
	 * ex. : <category ref="xx" type="collection">
				<name/>
				<description/>
			</category>
	 */
	class PolarisLinkImport_category extends PolarisLinkImportBase 
	{	
		private $Categories = null;
		private $CategoriesById = null;

		public $HandledCategoriesRoot = null;
		public $CategoryAssociationsDefaults = null;
		public $CategoryAssociations = null;
		public $CategoryAssociationsMarques = null;

		// -------------------------------- Gestion du mappage --------------------------
		protected function InitMapping()
		{
			// Les catégories gérées de Polaris
			$this->HandledCategoriesRoot = array(
				$this->Link->CategoryForRFS,
				$this->Link->CategoryForCollection,
				$this->Link->CategoryForSeason,
				$this->Link->CategoryForTags
			);

			if ($this->Link->CategoryForCategory1 != 0)
				$this->HandledCategoriesRoot[] = $this->Link->CategoryForCategory1;
			if ($this->Link->CategoryForCategory2 != 0)
				$this->HandledCategoriesRoot[] = $this->Link->CategoryForCategory2;
			if ($this->Link->CategoryForCategory3 != 0)
				$this->HandledCategoriesRoot[] = $this->Link->CategoryForCategory3;
			if ($this->Link->CategoryForCategory4 != 0)
				$this->HandledCategoriesRoot[] = $this->Link->CategoryForCategory4;

			// Chargement des associations
			$this->CategoryAssociations = [];
			$this->CategoryAssociationsMarques = [];
			$this->CategoryAssociationsDefaults = [];

			// Catégorie de Prestashop
			$psCategories = $this->GetCategoriesById();

			// Chargement des marques "nom" => "id" ...
			$manufacturers = [];
			foreach ($this->SafeSQL('SELECT id_manufacturer, name FROM `'._DB_PREFIX_.'manufacturer`') as $r)
				$manufacturers[strtolower(trim($r["name"]))] = $r["id_manufacturer"];

			// On cherche la catégorie représentant les marques		
			if (!empty($this->Link->CategoryForManufacturer) && isset($psCategories[$this->Link->CategoryForManufacturer]))
			{
				// On identifie toutes les marques des enfants de cette catégorie
				foreach ($psCategories[$this->Link->CategoryForManufacturer]["childs"] as $child)
				{
					// Chaque enfant est une marque à identifier
					$child_name = strtolower(trim($child["name"]));

					// Marque identifiée
					$id_manufacturer = isset($manufacturers[$child_name]) ? $manufacturers[$child_name] : "0";

					// Cette catégorie ainsi que tous ses enfants de cette catégorie reçoive la même marque
					$this->AssignManufacturerToCategoryAndItsChilds($psCategories, $child["id_category"], $id_manufacturer);
				}
			}
						
			foreach ($this->SafeSQL('SELECT id_category, id_category_l, is_default FROM `'._DB_PREFIX_.'polarislink_categories_map` WHERE id_link = '.$this->Link->LinkID) as $r)
			{
				$id_category_src = (string) $r["id_category"];
				$id_category_dst = (string) $r["id_category_l"];

				if (!isset($this->CategoryAssociations[$id_category_src]))
					$this->CategoryAssociations[$id_category_src] = [];
				
				$this->CategoryAssociations[$id_category_src][] = $id_category_dst;
				if ($r["is_default"])
					$this->CategoryAssociationsDefaults[$id_category_src] = $id_category_dst;
			}
		}

		/**
		 * On assigne la marque à une catégorie et ses enfants
		 */
		private function AssignManufacturerToCategoryAndItsChilds(&$psCategories, $id_category, $id_manufacturer)
		{			
			$this->CategoryAssociationsMarques[$id_category] = $id_manufacturer;
			if (isset($psCategories[$id_category]["childs"]))
			{
				foreach ($psCategories[$id_category]["childs"] as &$child)
					$this->AssignManufacturerToCategoryAndItsChilds($psCategories, $child["id_category"], $id_manufacturer);

				// Si pas d'enfant, alors c'est la catégorie par défaut pour les marques !!!
				if (count($psCategories[$id_category]["childs"]) == 0)
					$this->CategoryAssociationsDefaults[$id_category] = $id_category;
			}
		}

		/**
		 * Obtient les catégories actuelles du produit
		 */
		public function getCurrentProductCategories(Product $product)
		{
			$categories = [];
			foreach ($product->getCategories() as $category_id)
				$categories[(string) $category_id] = (string) $category_id;
			return $categories;			
		}

		/**
		 * Obtient la liste des ID de catégories Polaris gérées actuellement positionnées pour le produit $product
		 */
		public function getHandledCategoriesForProduct(Product $product)
		{
			// Association avec	les catégories ...
			$handled_categories = array();				

			foreach ($product->getCategories() as $category_id)
			{
				foreach ($this->HandledCategoriesRoot as $rootCategoryHandledByPolaris)
				{
					if ($this->isSubCategoryOf($rootCategoryHandledByPolaris, (string) $category_id))
						$handled_categories[(string) $category_id] = (string) $category_id;
				}
			}

			return $handled_categories;
		}

		/**
		 * Obtient la catégorie par défaut à positionner pour un produit
		 */
		public function computeDefaultCategoryForProduct(Product $product)
		{
			// On charge les catégories du produit
			$product_categories = $this->getCurrentProductCategories($product);
			$current_default_category_id = $product->id_category_default;

			// On vérifie que la catégorie par défaut existe et est bien positionnée pour le produit en question
			// si ce n'est pas le cas, on considère que le produit n'a pas de catégorie par défaut assignée
			if (!empty($current_default_category_id) && !isset($product_categories[$current_default_category_id]))
				$current_default_category_id = null;

			// On choisit notre mode d'affectation en fonction de la configuration
			if ($this->Link->AutoCategory)
			{
				$ps_categories = $this->GetCategoriesById();

				// Calcul des candidats C_ID => Score, on prendra le score le plus élevé
				$candidates = [];

				// Classement automatique
				if ( $this->Link->DefaultCategoryRoot != 0 && $this->Link->DefaultCategoryRoot != Configuration::get('PS_ROOT_CATEGORY') )
				{
					// Les catégories classiques
					foreach ($product_categories as $category_id)
					{
						$category_id = (string) $category_id;
						if (isset($this->CategoryAssociationsDefaults[$category_id]) && isset($ps_categories[$this->CategoryAssociationsDefaults[$category_id]]))
							$candidates[$this->CategoryAssociationsDefaults[$category_id]] = 1000 + $ps_categories[$this->CategoryAssociationsDefaults[$category_id]]["level"];

						if ($this->isSubCategoryOf($this->Link->DefaultCategoryRoot, $category_id))
							$candidates[$category_id] = $ps_categories[$category_id]["level"];
						
						if (!empty($this->Link->CategoryForManufacturer) && $this->isSubCategoryOf($this->Link->CategoryForManufacturer, $category_id))
							$candidates[$category_id] = 1000000 + $ps_categories[$category_id]["level"];
					}
				}
								
				// On tri par score et le vainqueur est :
				arsort($candidates);

				foreach ($candidates as $category_id => $ordre)
				{
					$current_default_category_id = (string) $category_id;
					break;
				}
			}
			else
			{
				// Classement manuel
				// Les produits sont dispatchés dans une catégorie "A classer" lorsqu'ils n'ont pas de catégorie déjà associée
				if (empty($current_default_category_id))
					$current_default_category_id = $this->Link->SortingCategoryID;
			}

			return $current_default_category_id;
		}

		/**
		 * Obtient un mapping pour un produit spécifique
		 * mapping : category_source => category_dest
		 */
		public function getMappingInfoForProduct(Product $product)
		{
			// On ne garde que les catégories d'associations qui sont compatibles avec la marque
			$mapping = [];
			foreach ($this->CategoryAssociations as $category_source_id => $mapped_categories)
			{
				foreach ($mapped_categories as $mapped_category_id)
				{
					if (!isset($this->CategoryAssociationsMarques[$mapped_category_id]) || 
							$this->CategoryAssociationsMarques[$mapped_category_id] == $product->id_manufacturer)
					{
						if (!isset($mapping[$category_source_id]))
							$mapping[$category_source_id] = [];

						$mapping[$category_source_id][] = $mapped_category_id;
					}
				}
			}

			return $mapping;
		}

		/** 
			Détermine si la catégorie $id est la descendante (directe ou non) de $category_id
		*/
		public function isSubCategoryOf($category_id, $id)
		{
			if ($category_id == $id)
				return true;
				
			$categories = $this->GetCategoriesById();
		
			if (empty($category_id) || !isset($categories[(string) $category_id]))
				return false;

			return in_array((string) $id, $categories[(string) $category_id]["descendancy"]);
		}

		// ------------------------------------------ Fin mapping
		
		/**
		 * Initialisation
		 */
		protected function OnInit()
		{
			parent::OnInit();
			
			$this->setCache(new PolarisLinkExtRefCache($this->Link->LinkID, "category"));
			$this->InitMapping();
		}
		
		public function GetCategories()
		{
			if ($this->Categories == null)
			{
				// Nous devons charger les catégories (pour aller plus vite)
				$this->Categories = array();
				
				foreach ($this->SafeSQL('
								SELECT c.id_category, c.id_parent, cl.name, c.active
								FROM `'._DB_PREFIX_.'category` c
								'.Shop::addSqlAssociation('category', 'c').'
								LEFT JOIN `'._DB_PREFIX_.'category_lang` cl ON (c.`id_category` = cl.`id_category`'.Shop::addSqlRestrictionOnLang('cl').' AND `id_lang` = '.$this->Link->LangID.")", true, false
							) as $category)
					$this->Categories[] = $category;
			}
			
			return $this->Categories;
		}
		
		public function GetCategoriesById()
		{
			if ($this->CategoriesById == null)
			{
				$this->CategoriesById = array();
				foreach ($this->GetCategories() as &$category)
					$this->CategoriesById[$category["id_category"]] = &$category;

				// On range les enfants dans une propriété childs
				foreach ($this->CategoriesById as $id => &$category)
				{
					$category["level"] = 0;			// Position dans l'arbre
					$category["childs"] = [];		// Liste des enfants immédiats
					$category["descendancy"] = []; 	// Liste des sous-catégories (enfants ou grand-enfants)
				}

				// On range les enfants dans une propriété childs
				foreach ($this->CategoriesById as $id => &$category)
					if (!empty($category["id_parent"]))
						$this->CategoriesById[$category["id_parent"]]["childs"][] = &$category;

				// Calcul de la descendance des racines
				foreach ($this->CategoriesById as $id => &$category)
				{
					if (empty($category["id_parent"]))
						$this->_ComputeDecendancy($category, 0);
				}
			}			

			return $this->CategoriesById;
		}

		/**
		 * Calcule la descendance des catégories
		 * 
		 */
		private function _ComputeDecendancy(&$category, $level)
		{
			if ($level > 100)
				return;	// cycle ??

			$category["level"] = $level;

			foreach ($category["childs"] as &$child)
			{
				$category["descendancy"][] = $child["id_category"];
				$this->_ComputeDecendancy($child, $level + 1);

				$category["descendancy"] = array_merge($category["descendancy"], $child["descendancy"]);
			}			
		}
		
		public function InvalidCache()
		{
			$this->Categories = null;
			$this->CategoriesById = null;
		}
		
		public function getParentCategoryId($data, $mainRoot = null, $parentTagName = "category", $parentType = null)
		{
			$parentId = $mainRoot;
			if (!empty($data->$parentTagName))
			{
				// Il y a un parent, on vérifie qu'il existe ...
				$psId = $this->ImportData($data->$parentTagName, $parentTagName, $parentType, $parentType);
				if (empty($psId))
					throw new Exception(sprintf("Impossible de trouver la racine de la catégorie (noeud cherché: %s)", $parentTagName));

				// Injection du parent
				if (isset($data["type"]) && $data["type"] == "c" && isset($data->$parentTagName["ref"]))
					$data->parent = (string) $data->$parentTagName["ref"];

				return $psId;
			}
			else
			{
				if ($mainRoot == NULL)
					throw new Exception("Aucun parent spécifié pour la catégorie alors qu'attendu");
			}
			
			return $mainRoot;
		}
		
		public function getCategory($polarisRef, $names, $parentId)
		{		
			$cache = $this->getCache("pr_category");
			return $cache->get($polarisRef);
		}
		
		public function getOrCreateCategory($polarisRef, $names, $parentId)
		{		
			$cache = $this->getCache("pr_category");
			$psId = $cache->get($polarisRef);

			// Vérification existence catégorie ...
			// Pas d'InnoDB ?
			if (!empty($psId))
			{
				$psCat = new Category($psId);
				if (empty($psCat->id))
				{
					$cache->set($polarisRef, null);
					$psId = null;
				}
			}
			
			if (empty($psId))
			{
				// On cherche la catégorie par son nom ...
				foreach ($names as $langId => $name)
				{
					$row = Category::searchByNameAndParentCategoryId($langId, $name, $parentId);
					if (!empty($row))
					{
						$psId = $row["id_category"];
						break;
					}
				}
				
				if (empty($psId))
				{
					$obj = new Category();					
					$obj->id_parent = $parentId;
					$obj->name = $this->AutoFillEmptyLangValue($names);
					$obj->link_rewrite = $this->BuildLinkRewrite($obj->name);					

					// référence non trouvée, création ...
					$parentCategory = new Category($parentId);
					$obj->active = $parentCategory->active;	
					
					$this->NewLogCtx($obj, "Création de la catégorie %s", [$obj->name[$this->Link->LangID]])->LogCreate()->Commit();
					
					$psId = $obj->id;
					$this->InvalidCache();
				}
				
				$cache->set($polarisRef, $psId);
			}
			
			return $psId;
		}
				
		protected function GetCategoryInfoForType($strType)
		{		
			switch ($strType)
			{
				case "ray": return array("CPR" => "ray %s", "root" => $this->Link->CategoryForRFS);
				case "family": return array("CPR" => "fam %s", "root" => NULL);
				case "rfs": return array("CPR" => "rfs %s", "root" => NULL);
				case "c": return array("CPR" => "c %s", "root" => $this->Link->CategoryForRFS);
				case "c1": return array("CPR" => "c1 %s", "root" => $this->Link->CategoryForCategory1);
				case "c2": return array("CPR" => "c2 %s", "root" => $this->Link->CategoryForCategory2);
				case "c3": return array("CPR" => "c3 %s", "root" => $this->Link->CategoryForCategory3);
				case "c4": return array("CPR" => "c4 %s", "root" => $this->Link->CategoryForCategory4);
				case "collection": return array("CPR" => "col %s", "root" => $this->Link->CategoryForCollection);
				case "tag": return array("CPR" => "tag %s", "root" => NULL);
				case "season": return array("CPR" => "ssn %s", "root" => $this->Link->CategoryForSeason);
				case "grtag": return array("CPR" => "gtg %s", "root" => $this->Link->CategoryForTags);
				case NULL:
				case "":
					return array("CPR" => "%s", "root" => (int)Configuration::get('PS_ROOT_CATEGORY'));
				default:
					throw new Exception(sprintf("Catégorie de type « %s » inconnue", $strType));
			}
		}
		
		protected function ConvertPolarisRef($categoryInfoType, $polarisRef)
		{
			return sprintf($categoryInfoType["CPR"], $polarisRef);
		}
		
		public function ImportData($data, $parentTagName = "category", $categoryType = null, $parentType = null)
		{
			if (!$this->Link->OPT_ALLOW_I_CATALOG)
				return;	// Rien à faire ici dans ce cas là !
			
			// On se positionne sur le bon shop !
			$this->SetCatalogContext();
			
			// Avons nous la référence Polaris ?
			$infoTypeCategory = $this->GetCategoryInfoForType(empty($categoryType) ? (string) $this->GetMandatory($data, "", "type") : $categoryType);
			$polarisRef = $this->ConvertPolarisRef(
								$infoTypeCategory,
								(string) $this->GetMandatory($data, "", "ref"));
								
			// Localisation de la marque
			$cache = $this->getCache("pr_category");
			$psId = $cache->get($polarisRef);
			// Activé ?
			$deleted = (string) $this->GetOptional($data, "", "deleted", "0");
			
			if (!$deleted)
			{			
				$obj = new Category($this->getOrCreateCategory($polarisRef, $this->getLgValue($data->name, "checkCategoryName"), 
							$this->getParentCategoryId($data, $infoTypeCategory["root"], $parentTagName, $parentType)));
			
				$log = $this->NewLogCtx($obj, "Modification de la catégorie %s", [$obj->name[$this->Link->LangID]]);
			
				if (count($data->name))
					$this->updProperty($log, $obj, "name", $data->name, 'checkCategoryName');
				
				if ($this->Link->OPT_ALLOW_I_DESCRIPTIONS && count($data->description) > 0)				
					$this->updOptProperty($log, $obj, "description", $data, "description");
				
				if ($this->Link->OPT_ALLOW_I_MDESCRIPTIONS && count($data->meta) > 0)
					$this->updOptProperty($log, $obj, "meta_description", $data, "meta");
				
				if ($data->parent && !empty((string) $data->parent))
				{
					$cache = $this->getCache("pr_category");
					$parentRef = (string) $data["type"] . " " . (string) $data->parent;
					$parentPsId = $cache->get($parentRef);
					if ($parentPsId && $obj->id_parent != $parentPsId)
					{
						$this->updProperty($log, $obj, "id_parent", $parentPsId);
						$this->InvalidCache();
					}
				}				

				// le pont ne doit pas réactiver pas des catégories masquées intentionnellement
				// $this->updProperty($log, $obj, "active", true);

				$log->Commit();
				
				if ($this->Link->OPT_ALLOW_I_PHOTOS && count($data->photos) > 0)
					$this->ImportPhotoLogo("catégorie", _PS_CAT_IMG_DIR_, "categories", "category", $obj, (string) $data->photos);
				
				$psId = $obj->id;
			}
			else if (!empty($psId))
			{		
				// Element non publié mais en cours sur le site, on laisse activé
				// tout comme on laisse les marques et autres catégories...
			}

			return $psId;
		}
		
		protected function OnImportData($data)
		{
			$this->ImportData($data);			
		}
	}
	